// Code generated by "stringer -type=LoadRestrictions"; DO NOT EDIT.

package types

import "strconv"

func _() {
	// An "invalid array index" compiler error signifies that the constant values have changed.
	// Re-run the stringer command to generate them again.
	var x [1]struct{}
	_ = x[LoadRestrictionsUnknown-0]
	_ = x[LoadRestrictionsRootOnly-1]
	_ = x[LoadRestrictionsNone-2]
}

const _LoadRestrictions_name = "LoadRestrictionsUnknownLoadRestrictionsRootOnlyLoadRestrictionsNone"

var _LoadRestrictions_index = [...]uint8{0, 23, 47, 67}

func (i LoadRestrictions) String() string {
	if i < 0 || i >= LoadRestrictions(len(_LoadRestrictions_index)-1) {
		return "LoadRestrictions(" + strconv.FormatInt(int64(i), 10) + ")"
	}
	return _LoadRestrictions_name[_LoadRestrictions_index[i]:_LoadRestrictions_index[i+1]]
}
