// Copyright 2014 Google Inc. All Rights Reserved.
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

//go:build linux

package install

import (
	"github.com/google/cadvisor/fs"
	"github.com/google/cadvisor/fs/nfs"

	"k8s.io/klog/v2"
)

func init() {
	err := fs.RegisterPlugin("nfs", nfs.NewPlugin())
	if err != nil {
		klog.Fatalf("Failed to register nfs fs plugin: %v", err)
	}
}
