//go:build !ignore_autogenerated
// +build !ignore_autogenerated

/*
Copyright The Kubernetes Authors.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/

// Code generated by validation-gen. DO NOT EDIT.

package v1

import (
	context "context"
	fmt "fmt"

	batchv1 "k8s.io/api/batch/v1"
	equality "k8s.io/apimachinery/pkg/api/equality"
	operation "k8s.io/apimachinery/pkg/api/operation"
	safe "k8s.io/apimachinery/pkg/api/safe"
	validate "k8s.io/apimachinery/pkg/api/validate"
	runtime "k8s.io/apimachinery/pkg/runtime"
	field "k8s.io/apimachinery/pkg/util/validation/field"
)

func init() { localSchemeBuilder.Register(RegisterValidations) }

// RegisterValidations adds validation functions to the given scheme.
// Public to allow building arbitrary schemes.
func RegisterValidations(scheme *runtime.Scheme) error {
	// type CronJob
	scheme.AddValidationFunc((*batchv1.CronJob)(nil), func(ctx context.Context, op operation.Operation, obj, oldObj interface{}) field.ErrorList {
		switch op.Request.SubresourcePath() {
		case "/":
			return Validate_CronJob(ctx, op, nil /* fldPath */, obj.(*batchv1.CronJob), safe.Cast[*batchv1.CronJob](oldObj))
		}
		return field.ErrorList{field.InternalError(nil, fmt.Errorf("no validation found for %T, subresource: %v", obj, op.Request.SubresourcePath()))}
	})
	return nil
}

// Validate_CronJob validates an instance of CronJob according
// to declarative validation rules in the API schema.
func Validate_CronJob(ctx context.Context, op operation.Operation, fldPath *field.Path, obj, oldObj *batchv1.CronJob) (errs field.ErrorList) {
	// field batchv1.CronJob.TypeMeta has no validation
	// field batchv1.CronJob.ObjectMeta has no validation

	// field batchv1.CronJob.Spec
	errs = append(errs,
		func(fldPath *field.Path, obj, oldObj *batchv1.CronJobSpec, oldValueCorrelated bool) (errs field.ErrorList) {
			// don't revalidate unchanged data
			if oldValueCorrelated && op.Type == operation.Update && equality.Semantic.DeepEqual(obj, oldObj) {
				return nil
			}
			// call the type's validation function
			errs = append(errs, Validate_CronJobSpec(ctx, op, fldPath, obj, oldObj)...)
			return
		}(fldPath.Child("spec"), &obj.Spec, safe.Field(oldObj, func(oldObj *batchv1.CronJob) *batchv1.CronJobSpec { return &oldObj.Spec }), oldObj != nil)...)

	// field batchv1.CronJob.Status has no validation
	return errs
}

// Validate_CronJobSpec validates an instance of CronJobSpec according
// to declarative validation rules in the API schema.
func Validate_CronJobSpec(ctx context.Context, op operation.Operation, fldPath *field.Path, obj, oldObj *batchv1.CronJobSpec) (errs field.ErrorList) {
	// field batchv1.CronJobSpec.Schedule
	errs = append(errs,
		func(fldPath *field.Path, obj, oldObj *string, oldValueCorrelated bool) (errs field.ErrorList) {
			// don't revalidate unchanged data
			if oldValueCorrelated && op.Type == operation.Update && (obj == oldObj || (obj != nil && oldObj != nil && *obj == *oldObj)) {
				return nil
			}
			// call field-attached validations
			earlyReturn := false
			if e := validate.RequiredValue(ctx, op, fldPath, obj, oldObj).MarkAlpha(); len(e) != 0 {
				errs = append(errs, e...)
				earlyReturn = true
			}
			if earlyReturn {
				return // do not proceed
			}
			return
		}(fldPath.Child("schedule"), &obj.Schedule, safe.Field(oldObj, func(oldObj *batchv1.CronJobSpec) *string { return &oldObj.Schedule }), oldObj != nil)...)

	// field batchv1.CronJobSpec.TimeZone has no validation
	// field batchv1.CronJobSpec.StartingDeadlineSeconds has no validation
	// field batchv1.CronJobSpec.ConcurrencyPolicy has no validation
	// field batchv1.CronJobSpec.Suspend has no validation
	// field batchv1.CronJobSpec.JobTemplate has no validation
	// field batchv1.CronJobSpec.SuccessfulJobsHistoryLimit has no validation
	// field batchv1.CronJobSpec.FailedJobsHistoryLimit has no validation
	return errs
}
