/*
Copyright 2017 The Kubernetes Authors.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/

package features

import (
	"k8s.io/apimachinery/pkg/util/runtime"
	"k8s.io/apimachinery/pkg/util/version"
	utilfeature "k8s.io/apiserver/pkg/util/feature"
	"k8s.io/component-base/featuregate"
)

// Every feature gate should have an entry here following this template:
//
// // owner: @username
// MyFeature() bool
//
// Feature gates should be listed in alphabetical, case-sensitive
// (upper before any lower case character) order. This reduces the risk
// of code conflicts because changes are more likely to be scattered
// across the file.
const (
	// owner: @michaelasp
	// kep: https://kep.k8s.io/4192
	//
	// Enables the tracking of observed generation in CRD status and conditions.
	CRDObservedGenerationTracking featuregate.Feature = "CRDObservedGenerationTracking"

	// owner: @alexzielenski
	//
	// Ignores errors raised on unchanged fields of Custom Resources
	// across UPDATE/PATCH requests.
	CRDValidationRatcheting featuregate.Feature = "CRDValidationRatcheting"

	// owner: @jpbetz
	//
	// CustomResourceDefinitions may include SelectableFields to declare which fields
	// may be used as field selectors.
	CustomResourceFieldSelectors featuregate.Feature = "CustomResourceFieldSelectors"
)

func init() {
	runtime.Must(utilfeature.DefaultMutableFeatureGate.AddVersioned(defaultVersionedKubernetesFeatureGates))
}

// defaultVersionedKubernetesFeatureGates consists of all known Kubernetes-specific feature keys with VersionedSpecs.
// To add a new feature, define a key for it above and add it below. The features will be
// available throughout Kubernetes binaries.
// To support n-3 compatibility version, features may only be removed 3 releases after graduation.
//
// Entries are alphabetized.
var defaultVersionedKubernetesFeatureGates = map[featuregate.Feature]featuregate.VersionedSpecs{
	CRDObservedGenerationTracking: {
		{Version: version.MustParse("1.35"), PreRelease: featuregate.Beta, Default: false},
		{Version: version.MustParse("1.36"), PreRelease: featuregate.Beta, Default: true},
	},
	CRDValidationRatcheting: {
		{Version: version.MustParse("1.28"), Default: false, PreRelease: featuregate.Alpha},
		{Version: version.MustParse("1.30"), Default: true, PreRelease: featuregate.Beta},
		{Version: version.MustParse("1.33"), Default: true, LockToDefault: true, PreRelease: featuregate.GA},
	},
	CustomResourceFieldSelectors: {
		{Version: version.MustParse("1.30"), Default: false, PreRelease: featuregate.Alpha},
		{Version: version.MustParse("1.31"), Default: true, PreRelease: featuregate.Beta},
		{Version: version.MustParse("1.32"), Default: true, LockToDefault: true, PreRelease: featuregate.GA},
	},
}
